@@ -1,54 +1,38 @@
 ##
 ##
 
-from .exceptions import (IndexNotReady, IndexNotFoundError, CollectionNameNotFound, IndexStatError, ClusterHealthCheckError)
+from __future__ import annotations
+from .cb_bucket import Bucket
+from .cb_index import CBQueryIndex
+from .exceptions import (IndexNotReady, IndexNotFoundError, CollectionNameNotFound, IndexStatError, ClusterHealthCheckError, PathMapUpsertError, CollectionUpsertError,
+                         ScopeCreateException, BucketCreateException, CollectionCreateException)
 from .retry import retry, retry_inline
 from .cb_connect import CBConnect
+from .util import r_getattr, omit_path, copy_path
+from .config import UpsertMapConfig, MapUpsertType
+from .cb_capella import Capella
+from .httpsessionmgr import APISession
 from datetime import timedelta
-import attr
 import hashlib
-from attr.validators import instance_of as io, optional
-from typing import Protocol, Iterable
+import logging
+import json
+import xmltodict
+import concurrent.futures
+from typing import Optional, Any
 from couchbase.cluster import Cluster
 from couchbase.options import QueryOptions
 from couchbase.diagnostics import ServiceType, PingState
-from couchbase.management.buckets import CreateBucketSettings, BucketType, StorageBackend
+from couchbase.management.buckets import CreateBucketSettings, BucketSettings
 from couchbase.management.collections import CollectionSpec
 from couchbase.exceptions import (QueryIndexNotFoundException, QueryIndexAlreadyExistsException, BucketAlreadyExistsException, BucketNotFoundException, BucketDoesNotExistException,
                                   WatchQueryIndexTimeoutException, ScopeAlreadyExistsException, CollectionAlreadyExistsException, CollectionNotFoundException)
 from couchbase.management.queries import (CreateQueryIndexOptions, CreatePrimaryQueryIndexOptions, WatchQueryIndexOptions, DropPrimaryQueryIndexOptions, DropQueryIndexOptions)
 from couchbase.management.options import CreateBucketOptions, CreateScopeOptions, CreateCollectionOptions, GetAllQueryIndexOptions
-from couchbase.options import WaitUntilReadyOptions
+from couchbase.options import WaitUntilReadyOptions, UpsertOptions
+from couchbase.management.logic.buckets_logic import BucketType, CompressionMode, ConflictResolutionType, EvictionPolicyType
 
-
-@attr.s
-class CBQueryIndex(Protocol):
-    name = attr.ib(validator=io(str))
-    is_primary = attr.ib(validator=io(bool))
-    state = attr.ib(validator=io(str))
-    namespace = attr.ib(validator=io(str))
-    keyspace = attr.ib(validator=io(str))
-    index_key = attr.ib(validator=io(Iterable))
-    condition = attr.ib(validator=io(str))
-    bucket_name = attr.ib(validator=optional(io(str)))
-    scope_name = attr.ib(validator=optional(io(str)))
-    collection_name = attr.ib(validator=optional(io(str)))
-    partition = attr.ib(validator=optional(validator=io(str)))
-
-    @classmethod
-    def from_server(cls, json_data):
-        return cls(json_data.get("name"),
-                   bool(json_data.get("is_primary")),
-                   json_data.get("state"),
-                   json_data.get("keyspace_id"),
-                   json_data.get("namespace_id"),
-                   json_data.get("index_key", []),
-                   json_data.get("condition", ""),
-                   json_data.get("bucket_id", json_data.get("keyspace_id", "")),
-                   json_data.get("scope_id", ""),
-                   json_data.get("keyspace_id", ""),
-                   json_data.get("partition", None)
-                   )
+logger = logging.getLogger('cbutil.manager')
+logger.addHandler(logging.NullHandler())
 
 
 class CBManager(CBConnect):
@@ -56,30 +40,87 @@ class CBManager(CBConnect):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
 
-    def create_bucket(self, name, quota: int = 256, replicas: int = 0):
-        self.logger.debug(f"create_bucket: create bucket {name}")
-        try:
-            bm = self._cluster.buckets()
-            bm.create_bucket(CreateBucketSettings(name=name,
-                                                  bucket_type=BucketType.COUCHBASE,
-                                                  storage_backend=StorageBackend.COUCHSTORE,
-                                                  num_replicas=replicas,
-                                                  ram_quota_mb=quota),
-                             CreateBucketOptions(timeout=timedelta(seconds=25)))
-        except BucketAlreadyExistsException:
-            pass
-        self.bucket(name)
+    def mgmt_api_post(self, endpoint, data):
+        s = APISession(self.username, self.password)
+        s.set_host(self.rally_host_name, self.ssl, self.admin_port)
+        response = s.api_post(endpoint, data)
+        return response
+
+    def mgmt_api_get(self, endpoint):
+        s = APISession(self.username, self.password)
+        s.set_host(self.rally_host_name, self.ssl, self.admin_port)
+        response = s.api_get(endpoint).json()
+        return response
+
+    def connect_cluster(self) -> CBManager:
+        self._cluster = self.session()
+        return self
+
+    def create_bucket(self, bucket: Bucket):
+        self._cluster.wait_until_ready(timedelta(seconds=5), WaitUntilReadyOptions(service_types=[ServiceType.KeyValue]))
+
+        result = self.get_bucket(bucket.name)
+        if result:
+            logger.debug(f"create_bucket: bucket {bucket.name} already exists")
+        else:
+            logger.debug(f"create_bucket: create bucket {bucket.name}")
+            if self.capella_project and self.capella_db:
+                project = Capella().get_project(self.capella_project)
+                if not project:
+                    raise BucketCreateException(f"Can not lookup Capella project {self.capella_project}")
+                project_id = project.get('id')
+                cluster = Capella(project_id=project_id).get_cluster(self.capella_db)
+                if not cluster:
+                    raise BucketCreateException(f"Can not find Capella database {self.capella_db}")
+                cluster_id = cluster.get('id')
+                logger.debug(f"Creating Capella bucket {bucket.name} in project {project_id} database {cluster_id}")
+                Capella(project_id=project_id).add_bucket(cluster_id, bucket)
+            else:
+                try:
+                    bm = self._cluster.buckets()
+                    # noinspection PyTypeChecker
+                    bm.create_bucket(CreateBucketSettings(
+                        name=bucket.name,
+                        flush_enabled=bucket.flush_enabled,
+                        replica_index=bucket.replica_index,
+                        ram_quota_mb=bucket.ram_quota_mb,
+                        num_replicas=bucket.num_replicas,
+                        bucket_type=BucketType(bucket.bucket_type.value),
+                        eviction_policy=EvictionPolicyType(bucket.eviction_policy.value),
+                        max_ttl=bucket.max_ttl,
+                        compression_mode=CompressionMode(bucket.compression_mode.value),
+                        conflict_resolution_type=ConflictResolutionType(bucket.conflict_resolution_type.value)
+                    ), CreateBucketOptions(timeout=timedelta(seconds=25)))
+                except BucketAlreadyExistsException:
+                    pass
+
+        self.bucket(bucket.name)
 
     def drop_bucket(self, name):
-        self.logger.debug(f"drop_bucket: drop bucket {name}")
+        logger.debug(f"drop_bucket: drop bucket {name}")
         try:
             bm = self._cluster.buckets()
             bm.drop_bucket(name)
         except (BucketNotFoundException, BucketDoesNotExistException):
             pass
 
+    def bucket_list_all(self):
+        bm = self._cluster.buckets()
+        return bm.get_all_buckets()
+
+    def scope_list_all(self):
+        cm = self._bucket.collections()
+        return cm.get_all_scopes()
+
+    @staticmethod
+    def collection_list_all(scope):
+        return scope.collections
+
     def create_scope(self, name):
-        self.logger.debug(f"create_scope: create scope {name}")
+        if not name:
+            raise ScopeCreateException(f"scope name can not be null")
+        self._cluster.wait_until_ready(timedelta(seconds=5), WaitUntilReadyOptions(service_types=[ServiceType.KeyValue]))
+        logger.debug(f"create_scope: create scope {name}")
         try:
             if name != "_default":
                 cm = self._bucket.collections()
@@ -88,11 +129,14 @@ class CBManager(CBConnect):
             pass
         self.scope(name)
 
-    def create_collection(self, name):
-        self.logger.debug(f"create_collection: create collection {name}")
+    def create_collection(self, name, max_ttl=0):
+        if not name:
+            raise CollectionCreateException(f"collection name can not be null")
+        self._cluster.wait_until_ready(timedelta(seconds=5), WaitUntilReadyOptions(service_types=[ServiceType.KeyValue]))
+        logger.debug(f"create_collection: create collection {name}")
         try:
             if name != "_default":
-                collection_spec = CollectionSpec(name, scope_name=self._scope.name)
+                collection_spec = CollectionSpec(name, scope_name=self._scope.name, max_ttl=timedelta(max_ttl))
                 cm = self._bucket.collections()
                 cm.create_collection(collection_spec, CreateCollectionOptions(timeout=timedelta(seconds=25)))
                 retry_inline(self.get_collection, cm, name)
@@ -100,6 +144,13 @@ class CBManager(CBConnect):
             pass
         self.collection(name)
 
+    def get_bucket(self, name: str) -> Optional[BucketSettings]:
+        try:
+            bm = self._cluster.buckets()
+            return bm.get_bucket(name)
+        except BucketDoesNotExistException:
+            return None
+
     @staticmethod
     def get_scope(cm, scope_name):
         return next((s for s in cm.get_all_scopes() if s.name == scope_name), None)
@@ -116,7 +167,7 @@ class CBManager(CBConnect):
 
     @retry()
     def drop_collection(self, name):
-        self.logger.debug(f"drop_collection: drop collection {name}")
+        logger.debug(f"drop_collection: drop collection {name}")
         try:
             collection_spec = CollectionSpec(name, scope_name=self._scope.name)
             cm = self._bucket.collections()
@@ -128,7 +179,7 @@ class CBManager(CBConnect):
         cluster = Cluster.connect(self.cb_connect_string, self.cluster_options)
         cluster.wait_until_ready(timedelta(seconds=30), WaitUntilReadyOptions(service_types=[ServiceType.Query, ServiceType.Management]))
 
-    @retry(factor=0.5)
+    @retry()
     def wait_for_index_ready(self):
         value = []
         query_str = r"SELECT * FROM system:indexes;"
@@ -208,7 +259,7 @@ class CBManager(CBConnect):
                     ]
                 }
             }
-            self.logger.debug(f"scanning bucket {b.name}")
+            logger.debug(f"scanning bucket {b.name}")
             bucket = cluster.bucket(b.name)
             cm = bucket.collections()
             scopes = cm.get_all_scopes()
@@ -217,10 +268,10 @@ class CBManager(CBConnect):
                     "name": s.name,
                     "collections": []
                 }
-                self.logger.debug(f"scanning scope {s.name}")
+                logger.debug(f"scanning scope {s.name}")
                 collections = s.collections
                 for c in collections:
-                    self.logger.debug(f"scanning collection {c.name}")
+                    logger.debug(f"scanning collection {c.name}")
                     primary_index = False
                     index_get_options = GetAllQueryIndexOptions(scope_name=s.name, collection_name=c.name)
                     indexes = qim.get_all_indexes(b.name, index_get_options)
@@ -266,7 +317,7 @@ class CBManager(CBConnect):
             index_options = CreatePrimaryQueryIndexOptions(deferred=False,
                                                            timeout=timedelta(seconds=timeout),
                                                            num_replicas=replica)
-        self.logger.debug(
+        logger.debug(
             f"cb_create_primary_index: creating primary index on {self._collection.name}")
         try:
             qim = self._cluster.query_indexes()
@@ -289,13 +340,42 @@ class CBManager(CBConnect):
         try:
             index_name = self.index_name(fields)
             qim = self._cluster.query_indexes()
-            self.logger.debug(
+            logger.debug(
                 f"creating index {index_name} on {','.join(fields)} for {self.keyspace}")
             qim.create_index(self._bucket.name, index_name, fields, index_options)
             return index_name
         except QueryIndexAlreadyExistsException:
             pass
 
+    @retry()
+    def cb_index_create(self, index: CBQueryIndex, timeout: int = 480):
+        if index.is_primary:
+            index_options = CreatePrimaryQueryIndexOptions()
+        else:
+            index_options = CreateQueryIndexOptions()
+
+        index_options.update(deferred=True)
+        index_options.update(timeout=timedelta(seconds=timeout))
+        index_options.update(num_replicas=index.num_replica)
+        index_options.update(ignore_if_exists=True)
+        if index.bucket_id:
+            index_options.update(scope_name=index.scope_id)
+            index_options.update(collection_name=index.keyspace_id)
+        if index.condition:
+            index_options.update(condition=index.condition)
+
+        if index.bucket_id:
+            bucket_name = index.bucket_id
+        else:
+            bucket_name = index.keyspace_id
+
+        qim = self._cluster.query_indexes()
+
+        if index.is_primary:
+            qim.create_primary_index(bucket_name, index_options)
+        else:
+            qim.create_index(bucket_name, index.name, index.index_key, index_options)
+
     @retry()
     def cb_drop_primary_index(self, timeout: int = 120):
         if self._collection_name != '_default':
@@ -304,7 +384,7 @@ class CBManager(CBConnect):
                                                          scope_name=self._scope.name)
         else:
             index_options = DropPrimaryQueryIndexOptions(timeout=timedelta(seconds=timeout))
-        self.logger.debug(f"cb_drop_primary_index: dropping primary index on {self.collection_name}")
+        logger.debug(f"cb_drop_primary_index: dropping primary index on {self.collection_name}")
         try:
             qim = self._cluster.query_indexes()
             qim.drop_primary_index(self._bucket.name, index_options)
@@ -320,7 +400,7 @@ class CBManager(CBConnect):
         else:
             index_options = DropQueryIndexOptions(timeout=timedelta(seconds=timeout))
         try:
-            self.logger.debug(f"cb_drop_index: drop index {name}")
+            logger.debug(f"cb_drop_index: drop index {name}")
             qim = self._cluster.query_indexes()
             qim.drop_index(self._bucket.name, name, index_options)
         except QueryIndexNotFoundException:
@@ -334,7 +414,7 @@ class CBManager(CBConnect):
 
         for row in results:
             for key, value in row.items():
-                entry = CBQueryIndex.from_server(value)
+                entry = CBQueryIndex.from_dict(value)
                 all_list.append(entry)
 
         return all_list
@@ -346,7 +426,7 @@ class CBManager(CBConnect):
             index_list = self.index_list_all()
             for item in index_list:
                 if index_name == '#primary':
-                    if (item.collection_name == self.collection_name or item.bucket_name == self.collection_name) \
+                    if (item.keyspace_id == self.collection_name or item.keyspace_id == self._bucket_name) \
                             and item.name == '#primary':
                         return True
                 elif item.name == index_name:
@@ -356,7 +436,7 @@ class CBManager(CBConnect):
 
         return False
 
-    @retry(factor=0.5, allow_list=(IndexNotReady,))
+    @retry(allow_list=(IndexNotReady,))
     def index_wait(self, index_name: str = None):
         record_count = self.collection_count()
         try:
@@ -372,7 +452,7 @@ class CBManager(CBConnect):
 
         for item in index_list:
             if item.name == index_name and (
-                    item.collection_name == self.collection_name or item.bucket_name == self.collection_name):
+                    item.keyspace_id == self.collection_name or item.keyspace_id == self._bucket_name):
                 if len(list(item.index_key)) == 0:
                     return doc_key_field
                 else:
@@ -413,19 +493,174 @@ class CBManager(CBConnect):
         except WatchQueryIndexTimeoutException:
             raise IndexNotReady(f"Indexes not build within {timeout} seconds...")
 
-    @retry(factor=0.5, allow_list=(IndexNotReady,))
+    @retry(allow_list=(IndexNotReady,))
     def index_list(self):
         return_list = {}
         try:
             index_list = self.index_list_all()
             for item in index_list:
-                if item.collection_name == self.collection_name or item.bucket_name == self.collection_name:
+                if item.keyspace_id == self.collection_name or item.keyspace_id == self._bucket_name:
                     return_list[item.name] = item.state
             return return_list
         except Exception as err:
             raise IndexNotReady(f"index_list: bucket {self._bucket.name} error: {err}")
 
-    @retry(factor=0.5, allow_list=(IndexNotReady,))
+    @retry(allow_list=(IndexNotReady,))
     def delete_wait(self, index_name: str = None):
         if self.is_index(index_name=index_name):
             raise IndexNotReady(f"delete_wait: index still exists")
+
+    def cb_map_upsert_attr(self, prefix: str, config: UpsertMapConfig, attr_obj: Any):
+        primitive = (int, str, bool, list, dict)
+        value = None
+        for c in config.paths:
+            logger.debug(f"cb_map_upsert: processing key {c.path} name {c.name}")
+            try:
+                subset = r_getattr(attr_obj, c.path)
+                if isinstance(subset, list):
+                    if not any(isinstance(item, primitive) for item in subset):
+                        value = []
+                        for item in subset:
+                            value.append(item.as_dict)
+                elif isinstance(subset, primitive) or not subset:
+                    value = subset
+                else:
+                    value = subset.as_dict
+
+                if c.collection:
+                    self.create_collection(c.name)
+
+                if c.exclude:
+                    logger.debug(f"cb_map_upsert: excluding {','.join(c.exclude)}")
+                    value = omit_path(value, c.exclude)
+
+                if c.p_type == MapUpsertType.DOCUMENT:
+                    doc_id = self.key_format(c.id, value, text=prefix)
+                    logger.debug(f"cb_map_upsert: processing doc ID {doc_id}")
+                    self._collection.upsert(doc_id, {c.name: value})
+                elif c.p_type == MapUpsertType.LIST:
+                    logger.debug(f"cb_map_upsert: processing list")
+                    if not isinstance(value, list):
+                        raise PathMapUpsertError(f"cb_map_upsert: path {c.path} type {type(value)} incompatible with list mode")
+                    for doc in value:
+                        doc_id = self.key_format(c.id, doc, text=prefix, id_key=c.id_key)
+                        self._collection.upsert(doc_id, doc)
+
+            except AttributeError:
+                raise PathMapUpsertError(f"cb_map_upsert: key {c.path} not found")
+
+            except Exception as err:
+                raise PathMapUpsertError(f"cb_map_upsert: error {err}")
+
+    @retry()
+    def _cb_upsert(self, cluster, c_name, meta_id, doc_data, timeout=5):
+        upsert_options = UpsertOptions(timeout=timedelta(seconds=timeout))
+        try:
+            logger.debug(f"upsert -> {c_name}: {meta_id}")
+            bucket = cluster.bucket(self._bucket_name)
+            collection = bucket.scope(self._scope_name).collection(c_name)
+            result = collection.upsert(meta_id, doc_data, upsert_options)
+            return result.cas
+        except Exception as error:
+            raise CollectionUpsertError(f"upsert error: {error}")
+
+    @retry()
+    def _create_collection(self, cluster, c_name):
+        bucket = cluster.bucket(self._bucket_name)
+        cm = bucket.collections()
+        try:
+            collection_spec = CollectionSpec(c_name, scope_name=self._scope_name)
+            cm.create_collection(collection_spec, CreateCollectionOptions(timeout=timedelta(seconds=10)))
+            self._verify_collection(cm, c_name)
+        except CollectionAlreadyExistsException:
+            pass
+        except Exception as error:
+            raise CollectionCreateException(f"collection create error: {error}")
+
+    @retry()
+    def _verify_collection(self, cm, c_name):
+        sm = next((s for s in cm.get_all_scopes() if s.name == self._scope_name), None)
+        valid = next((i for i in sm.collections if i.name == c_name), None)
+        if not valid:
+            raise CollectionCreateException(f"collection {c_name} was not created")
+
+    def cb_map_upsert(self,
+                      prefix: str,
+                      config: UpsertMapConfig,
+                      json_file: str = None,
+                      xml_file: str = None,
+                      json_data: str = None,
+                      xml_data: str = None,
+                      timeout=5):
+        tasks = set()
+        executor = concurrent.futures.ThreadPoolExecutor()
+        cluster = Cluster.connect(self.cb_connect_string, self.cluster_options)
+
+        if json_file:
+            with open(json_file, mode="r") as json_xml:
+                data = json.load(json_xml)
+        elif xml_file:
+            with open(xml_file, mode="rb") as input_xml:
+                contents = input_xml.read()
+                data = xmltodict.parse(contents)
+        elif json_data:
+            data = json.loads(json_data)
+        elif xml_data:
+            data = xmltodict.parse(xml_data)
+        else:
+            raise PathMapUpsertError(f"cb_map_upsert: JSON or XML input data is required")
+
+        for c in config.paths:
+            if c.collection:
+                logger.debug(f"cb_map_upsert: creating collection {c.name}")
+                tasks.add(executor.submit(self._create_collection, cluster, c.name))
+
+        while tasks:
+            done, tasks = concurrent.futures.wait(tasks, return_when=concurrent.futures.FIRST_COMPLETED)
+            for task in done:
+                try:
+                    task.result()
+                except Exception as err:
+                    raise PathMapUpsertError(f"cb_map_upsert: {err}")
+
+        tasks.clear()
+        for c in config.paths:
+            logger.debug(f"cb_map_upsert: processing key {c.path} name {c.name}")
+
+            subset = copy_path(c.path, data)
+
+            if not subset or len(subset) == 0:
+                if c.optional:
+                    continue
+                else:
+                    raise PathMapUpsertError(f"path {c.path} not found in source data")
+
+            if c.collection:
+                collection_name = c.name
+            else:
+                collection_name = self._collection.name
+
+            if c.exclude:
+                logger.debug(f"cb_map_upsert: excluding {','.join(c.exclude)}")
+                subset = omit_path(subset, c.exclude)
+
+            if c.p_type == MapUpsertType.DOCUMENT:
+                doc_id = self.key_format(c.id, subset, text=prefix)
+                logger.debug(f"cb_map_upsert: processing doc ID {doc_id}")
+                doc = {c.name: subset}
+                tasks.add(executor.submit(self._cb_upsert, cluster, collection_name, doc_id, doc, timeout))
+            elif c.p_type == MapUpsertType.LIST:
+                logger.debug(f"cb_map_upsert: processing list")
+                if not isinstance(subset, list):
+                    raise PathMapUpsertError(f"cb_map_upsert: path {c.path} type {type(subset)} incompatible with list mode")
+                for doc in subset:
+                    doc_id = self.key_format(c.id, doc, text=prefix, id_key=c.id_key)
+                    tasks.add(executor.submit(self._cb_upsert, cluster, collection_name, doc_id, doc, timeout))
+
+        while tasks:
+            done, tasks = concurrent.futures.wait(tasks, return_when=concurrent.futures.FIRST_COMPLETED)
+            for task in done:
+                try:
+                    task.result()
+                except Exception as err:
+                    raise PathMapUpsertError(f"cb_map_upsert: {err}")
