@@ -1,33 +1,55 @@
 ##
 ##
 
-from .exceptions import (DNSLookupTimeout, NodeUnreachable, NodeConnectionTimeout, NodeConnectionError, NodeConnectionFailed, ClusterHealthCheckError)
+from .exceptions import (DNSLookupTimeout, NodeUnreachable, NodeConnectionTimeout, NodeConnectionError, NodeConnectionFailed, ClusterHealthCheckError, KeyFormatError)
 from .retry import retry
 from .httpsessionmgr import APISession
+from .config import KeyStyle
 import logging
 import socket
 import dns.resolver
+import uuid
+from typing import Union
+from enum import Enum
 from datetime import timedelta
 from couchbase.auth import PasswordAuthenticator
-from couchbase.options import ClusterTimeoutOptions, ClusterOptions, LockMode
+from couchbase.options import ClusterTimeoutOptions, LockMode, ClusterOptions, TLSVerifyMode
 from couchbase.cluster import Cluster
+from acouchbase.cluster import AsyncCluster
 from couchbase.diagnostics import ServiceType, PingState
 
+logger = logging.getLogger('cbutil.session')
+logger.addHandler(logging.NullHandler())
+
+
+class BucketMode(Enum):
+    DEFAULT = 0
+    CACHE = 1
+    CAPACITY = 2
+
 
 class CBSession(object):
 
-    def __init__(self, hostname: str, username: str, password: str, ssl=False, external=False):
+    def __init__(self, hostname: str, username: str, password: str, ssl=False, project=None, database=None, external=False, kv_timeout: int = 5, query_timeout: int = 60):
         self.cluster_node_count = None
-        self.logger = logging.getLogger(self.__class__.__name__)
         self._cluster = None
         self._bucket = None
         self._scope = None
         self._collection = None
+        self._bucket_name = None
         self._scope_name = "_default"
         self._collection_name = "_default"
+        self.kv_timeout = kv_timeout
+        self.query_timeout = query_timeout
+        self.hostname = hostname
         self.username = username
         self.password = password
-        self.ssl = ssl
+        self.capella_project = project
+        self.capella_db = database
+        if self.capella_db:
+            self.ssl = True
+        else:
+            self.ssl = ssl
         self.rally_host_name = hostname
         self.rally_cluster_node = self.rally_host_name
         self.rally_dns_domain = False
@@ -42,13 +64,18 @@ class CBSession(object):
         self.sw_version = None
         self.memory_quota = None
         self.cluster_services = []
+        self.memory_quota = 0
+        self.index_memory_quota = 0
+        self.fts_memory_quota = 0
+        self.cbas_memory_quota = 0
+        self.eventing_memory_quota = 0
         self.auth = PasswordAuthenticator(self.username, self.password)
-        self.timeouts = ClusterTimeoutOptions(query_timeout=timedelta(seconds=60),
-                                              kv_timeout=timedelta(seconds=4),
-                                              bootstrap_timeout=timedelta(seconds=4),
-                                              resolve_timeout=timedelta(seconds=4),
-                                              connect_timeout=timedelta(seconds=4),
-                                              management_timeout=timedelta(seconds=4))
+        self.timeouts = ClusterTimeoutOptions(query_timeout=timedelta(seconds=query_timeout),
+                                              kv_timeout=timedelta(seconds=kv_timeout),
+                                              bootstrap_timeout=timedelta(seconds=kv_timeout),
+                                              resolve_timeout=timedelta(seconds=kv_timeout),
+                                              connect_timeout=timedelta(seconds=kv_timeout),
+                                              management_timeout=timedelta(seconds=kv_timeout))
 
         if self.ssl:
             self.prefix = "https://"
@@ -63,8 +90,31 @@ class CBSession(object):
             self.admin_port = "8091"
             self.node_port = "9102"
 
+        self.is_reachable()
+        self.check_cluster()
+
+        self.cluster_options = ClusterOptions(self.auth,
+                                              timeout_options=self.timeouts,
+                                              tls_verify=TLSVerifyMode.NO_VERIFY,
+                                              lockmode=LockMode.WAIT)
+
+        if self.use_external_network:
+            self.cluster_options.update(network="external")
+        else:
+            self.cluster_options.update(network="default")
+
+    @retry()
+    def session(self) -> Cluster:
+        return Cluster.connect(self.cb_connect_string, self.cluster_options)
+
+    @retry()
+    async def session_a(self) -> AsyncCluster:
+        cluster = await AsyncCluster.connect(self.cb_connect_string, self.cluster_options)
+        await cluster.on_connect()
+        return cluster
+
     def construct_key(self, key):
-        if type(key) == int or str(key).isdigit():
+        if type(key) is int or str(key).isdigit():
             if self._collection.name != "_default":
                 return self._collection.name + ':' + str(key)
             else:
@@ -89,16 +139,16 @@ class CBSession(object):
     @property
     def cb_connect_string(self):
         connect_string = self.cb_prefix + self.rally_host_name
-        self.logger.debug(f"Connect string: {connect_string}")
+        logger.debug(f"Connect string: {connect_string}")
         return connect_string
 
-    @retry(retry_count=5)
+    @retry(retry_count=7)
     def is_reachable(self):
         resolver = dns.resolver.Resolver()
         resolver.timeout = 5
         resolver.lifetime = 10
 
-        self.logger.debug(f"checking if rally node is reachable: {self.rally_host_name}")
+        logger.debug(f"checking if rally node is reachable: {self.rally_host_name}")
         try:
             answer = resolver.resolve(self.srv_prefix + self.rally_host_name, "SRV")
             for srv in answer:
@@ -117,7 +167,7 @@ class CBSession(object):
 
         if self.rally_dns_domain:
             self.rally_cluster_node = self.rally_host_name = self.srv_host_list[0]['hostname']
-            self.logger.debug(f"Rewriting rally node as {self.rally_cluster_node}")
+            logger.debug(f"Rewriting rally node as {self.rally_cluster_node}")
 
         try:
             self.check_node_connectivity(self.rally_cluster_node, self.admin_port)
@@ -136,7 +186,7 @@ class CBSession(object):
         rally_ip = socket.gethostbyname(self.rally_host_name)
 
         if not self.cluster_info:
-            self.logger.debug("process_cluster_data: no cluster info")
+            logger.debug("process_cluster_data: no cluster info")
             return
 
         self.cluster_node_count = range(len(self.cluster_info['nodes']))
@@ -150,10 +200,16 @@ class CBSession(object):
                 self.external_list.append(alternate_address)
                 external_ip = socket.gethostbyname(alternate_address)
                 if rally_ip == external_ip:
-                    self.logger.debug(f"external address {rally_ip} detected")
+                    logger.debug(f"external address {rally_ip} detected")
                     self.use_external_network = True
 
-    @retry(retry_count=5)
+        self.memory_quota = self.cluster_info.get('memoryQuota', 0)
+        self.index_memory_quota = self.cluster_info.get('indexMemoryQuota', 0)
+        self.fts_memory_quota = self.cluster_info.get('ftsMemoryQuota', 0)
+        self.cbas_memory_quota = self.cluster_info.get('cbasMemoryQuota', 0)
+        self.eventing_memory_quota = self.cluster_info.get('eventingMemoryQuota', 0)
+
+    @retry(retry_count=7)
     def check_node_connectivity(self, hostname, port):
         try:
             sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
@@ -170,13 +226,13 @@ class CBSession(object):
         else:
             raise NodeConnectionFailed(f"node {hostname}:{port} unreachable")
 
-    @retry(factor=0.5)
+    @retry()
     def wait_until_ready(self):
         nodes = []
         cluster = Cluster(self.cb_connect_string, ClusterOptions(self.auth,
                                                                  timeout_options=self.timeouts,
                                                                  lockmode=LockMode.WAIT))
-        self.logger.debug(f"cluster {self.cb_connect_string} ping")
+        logger.debug(f"cluster {self.cb_connect_string} ping")
         ping_result = cluster.ping()
         endpoint: ServiceType
         for endpoint, reports in ping_result.endpoints.items():
@@ -187,7 +243,7 @@ class CBSession(object):
                     raise ClusterHealthCheckError(f"service {endpoint.value} not ok")
 
         node_set = set(nodes)
-        self.logger.debug("ping complete")
+        logger.debug("ping complete")
         return list(node_set)
 
     def print_host_map(self):
@@ -215,3 +271,59 @@ class CBSession(object):
                 for key in ext_port_list:
                     print("%s:%s" % (key, ext_port_list[key]), end=' ')
             print("[Services] %s [version] %s [platform] %s" % (services, version, ostype))
+
+    def get_quota_settings(self):
+        return dict(
+            data=self.memory_quota,
+            index=self.index_memory_quota,
+            fts=self.fts_memory_quota,
+            analytics=self.cbas_memory_quota,
+            eventing=self.eventing_memory_quota
+        )
+
+    def key_format(self,
+                   style: KeyStyle,
+                   document: dict,
+                   doc_num: int = 1,
+                   id_key: str = "record_id",
+                   separator: str = "::",
+                   text: Union[str, None] = None,
+                   path: Union[str, None] = None,
+                   field: Union[str, None] = None):
+        if style.value == 0:
+            return f"{self.keyspace}{separator}{doc_num}"
+        elif style.value == 1:
+            if not document.get('type'):
+                raise KeyFormatError(f"Key style type requested: document does not have type field")
+            if document.get(id_key):
+                number = document.get(id_key)
+            else:
+                number = doc_num
+            return f"{document['type']}{separator}{number}"
+        elif style.value == 2:
+            return uuid.uuid4()
+        elif style.value == 3:
+            if not field:
+                raise KeyFormatError(f"Key field name style requested: field parameter is null")
+            return f"{field}{separator}{doc_num}"
+        elif style.value == 4:
+            return f"{self.keyspace}{separator}{doc_num}"
+        elif style.value == 5:
+            if not field:
+                raise KeyFormatError(f"Key compound name style requested: field parameter is null")
+            return f"{self.keyspace}{separator}{field}{separator}{doc_num}"
+        elif style.value == 6:
+            if not path or not text:
+                raise KeyFormatError(f"Key path style parameters not provided")
+            return f"{text}{separator}{path}"
+        elif style.value == 7:
+            if not text:
+                raise KeyFormatError(f"Key text style parameter not provided")
+            return f"{text}"
+        elif style.value == 8:
+            if not text:
+                raise KeyFormatError(f"Key text style parameter not provided")
+            number = document.get(id_key, 1)
+            return f"{text}{separator}{id_key}{separator}{number}"
+        else:
+            raise KeyFormatError(f"Unknown key style {style.name}")
