@@ -11,9 +11,16 @@ import datetime
 import hmac
 import hashlib
 import warnings
+from enum import Enum
 from urllib.parse import urlparse
 from requests.auth import AuthBase
-from .exceptions import (NotAuthorized, HTTPForbidden, HTTPNotImplemented, RequestValidationError, InternalServerError, PaginationDataNotFound, SyncGatewayOperationException)
+from .exceptions import (NotAuthorized, HTTPForbidden, HTTPNotImplemented, RequestValidationError, InternalServerError, APIError,
+                         PaginationDataNotFound, SyncGatewayOperationException, PreconditionFailed, ConflictException, BadRequest)
+
+
+class AuthType(Enum):
+    basic = 0
+    capella = 1
 
 
 class CapellaToken(object):
@@ -59,19 +66,31 @@ class CapellaToken(object):
 class CapellaAuth(AuthBase):
 
     def __init__(self):
-        if 'CBC_ACCESS_KEY' in os.environ:
-            self.capella_key = os.environ['CBC_ACCESS_KEY']
-        else:
-            raise Exception("Please set CBC_ACCESS_KEY for Capella API access")
+        _credential_file = os.path.join(os.environ['HOME'], '.capella', 'default-api-key-token.txt')
+        _profile_token = None
+        self.profile_token = None
 
-        if 'CBC_SECRET_KEY' in os.environ:
-            self.capella_secret = os.environ['CBC_SECRET_KEY']
+        if os.path.exists(_credential_file):
+            try:
+                credential_data = dict(line.split(':', 1) for line in open(_credential_file))
+                _profile_token = credential_data.get('APIKeyToken')
+                if _profile_token:
+                    _profile_token = _profile_token.strip()
+            except Exception as err:
+                raise Exception(f"can not read credential file {_credential_file}: {err}")
+
+        if 'CAPELLA_TOKEN' in os.environ:
+            self.profile_token = os.environ['CAPELLA_TOKEN']
+        elif _profile_token:
+            self.profile_token = _profile_token
         else:
-            raise Exception("Please set CBC_SECRET_KEY for Capella API access")
+            raise Exception("Please set Capella Token for Capella API access (for example in $HOME/.capella/default-api-key-token.txt)")
 
     def __call__(self, r):
-        cbc_api_request_headers = CapellaToken(self.capella_key, self.capella_secret).signature(r.method, r.url).token
-        r.headers.update(cbc_api_request_headers)
+        request_headers = {
+            "Authorization": f"Bearer {self.profile_token}",
+        }
+        r.headers.update(request_headers)
         return r
 
 
@@ -98,19 +117,20 @@ class APISession(object):
     AUTH_BASIC = 0
     AUTH_CAPELLA = 1
 
-    def __init__(self, username=None, password=None, auth_type=0):
+    def __init__(self, username=None, password=None, auth_type=AuthType.basic):
         warnings.filterwarnings("ignore")
         self.username = username
         self.password = password
+        self.timeout = 30
         self.logger = logging.getLogger(self.__class__.__name__)
         self.url_prefix = "http://127.0.0.1"
         self.session = requests.Session()
-        retries = Retry(total=60,
-                        backoff_factor=0.2)
+        retries = Retry(total=10,
+                        backoff_factor=0.01)
         self.session.mount('http://', HTTPAdapter(max_retries=retries))
         self.session.mount('https://', HTTPAdapter(max_retries=retries))
         self._response = None
-        if auth_type == 0:
+        if auth_type == AuthType.basic:
             self.auth_class = BasicAuth(self.username, self.password)
         else:
             self.auth_class = CapellaAuth()
@@ -137,14 +157,20 @@ class APISession(object):
 
     def check_status_code(self, code):
         self.logger.debug("API status code {}".format(code))
-        if code == 200 or code == 201:
+        if code == 200 or code == 201 or code == 202 or code == 204:
             return True
+        elif code == 400:
+            raise BadRequest("Bad Request")
         elif code == 401:
             raise NotAuthorized("API: Unauthorized")
         elif code == 403:
             raise HTTPForbidden("API: Forbidden: Insufficient privileges")
         elif code == 404:
             raise HTTPNotImplemented("API: Not Found")
+        elif code == 409:
+            raise ConflictException("Conflict")
+        elif code == 412:
+            raise PreconditionFailed("Precondition Failed")
         elif code == 415:
             raise RequestValidationError("API: invalid body contents")
         elif code == 422:
@@ -164,6 +190,9 @@ class APISession(object):
             port_num = port if port else 443
             self.url_prefix = f"https://{hostname}:{port_num}"
 
+    def set_timeout(self, timeout: int):
+        self.timeout = timeout
+
     def get_endpoint(self, path):
         return ':'.join(self.url_prefix.split(':')[:-1]) + path
 
@@ -203,10 +232,7 @@ class APISession(object):
     def capella_pagination(response_json):
         if "cursor" in response_json:
             if "pages" in response_json["cursor"]:
-                if "items" in response_json["data"]:
-                    data = response_json["data"]["items"]
-                else:
-                    data = response_json["data"]
+                data = response_json["data"]
                 if "next" in response_json["cursor"]["pages"]:
                     next_page = response_json["cursor"]["pages"]["next"]
                     per_page = response_json["cursor"]["pages"]["perPage"]
@@ -219,7 +245,7 @@ class APISession(object):
     def api_get(self, endpoint, items=None):
         if items is None:
             items = []
-        response = self.session.get(self.url_prefix + endpoint, auth=self.auth_class, verify=False, timeout=15)
+        response = self.session.get(self.url_prefix + endpoint, auth=self.auth_class, verify=False, timeout=self.timeout)
 
         try:
             self.check_status_code(response.status_code)
@@ -240,17 +266,30 @@ class APISession(object):
         self._response = response_text
         return self
 
+    def capella_get(self, endpoint):
+        page = 1
+        items = []
+        while True:
+            url = f"{self.url_prefix}{endpoint}?page={page}&perPage=100"
+            response = self.session.get(url, auth=self.auth_class, verify=False, timeout=self.timeout)
+            response_json = json.loads(response.text)
+            items.extend(response_json.get('data'))
+            page = response_json.get('cursor', {}).get('pages', {}).get('next', 0)
+            if page == 0:
+                break
+        return items
+
     def api_post(self, endpoint, body):
         response = self.session.post(self.url_prefix + endpoint,
                                      auth=self.auth_class,
                                      json=body,
                                      verify=False,
-                                     timeout=15)
+                                     timeout=self.timeout)
 
         try:
             self.check_status_code(response.status_code)
-        except Exception:
-            raise
+        except Exception as err:
+            raise APIError(err, response.text, response.status_code) from err
 
         self._response = response.text
         return self
@@ -260,7 +299,7 @@ class APISession(object):
                                     auth=self.auth_class,
                                     json=body,
                                     verify=False,
-                                    timeout=15)
+                                    timeout=self.timeout)
 
         try:
             self.check_status_code(response.status_code)
@@ -277,7 +316,7 @@ class APISession(object):
                                     auth=self.auth_class,
                                     data=body,
                                     verify=False,
-                                    timeout=15,
+                                    timeout=self.timeout,
                                     headers=headers)
 
         try:
@@ -289,7 +328,7 @@ class APISession(object):
         return self
 
     def api_delete(self, endpoint):
-        response = self.session.delete(self.url_prefix + endpoint, auth=self.auth_class, verify=False, timeout=15)
+        response = self.session.delete(self.url_prefix + endpoint, auth=self.auth_class, verify=False, timeout=self.timeout)
 
         try:
             self.check_status_code(response.status_code)
@@ -298,3 +337,18 @@ class APISession(object):
 
         self._response = response.text
         return self
+
+    def api_patch(self, endpoint, body):
+        response = self.session.patch(self.url_prefix + endpoint,
+                                      auth=self.auth_class,
+                                      json=body,
+                                      verify=False,
+                                      timeout=self.timeout)
+
+        try:
+            self.check_status_code(response.status_code)
+        except Exception as err:
+            raise APIError(err, response.text, response.status_code) from err
+
+        self._response = response.text
+        return self
